import asyncio
import base64
import io
import json
from logging import exception
import time
from hashlib import md5
from urllib.parse import urlencode

import httpx
import qrcode
from httpx import ConnectTimeout, ReadTimeout
from nonebot.log import logger

from ..database import DB as Config


class RequestError(Exception):
    def __init__(self, code, message=None, data=None):
        self.code = code
        self.message = message
        self.data = data
    
    def __repr__(self):
        return f"<RequestError code={self.code} message={self.message}>"
    
    def __str__(self):
        return self.__repr__()


class BiliReq():
    def __init__(self):
        self.appkey = "4409e2ce8ffd12b8"
        self.appsec = "59b43e04ad6965f34319062b478f83dd"
        self.default_headers = {
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64)\
             AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88\
              Safari/537.36 Edg/87.0.664.60',
            'Referer': 'https://www.bilibili.com/'
        }
        self.login = Config.get_login()
        self.proxies = {
            'http': None,
            'https': None
        }

    async def request(self, method, url, **kw):
        async with httpx.AsyncClient(proxies=self.proxies) as client:
            try:
                res = await client.request(method, url, **kw)
                res.encoding = 'utf-8'
                res = res.json()
            except ConnectTimeout:
                logger.error(f"连接超时（{url}）")
            except ReadTimeout:
                logger.error(f"接收超时（{url}）")
                raise
            except exception as e:
                logger.error(f"未知错误（url）")
                raise 
            
            if res['code'] != 0:
                raise RequestError(code=res['code'],
                                    message=res['message'],
                                    data=res['data'])
            return res['data']
    
    async def get(self, url, **kw):
        return await self.request('GET', url, **kw)

    async def post(self, url, **kw):
        return await self.request('POST', url, **kw)
    
    async def get_info(self, uid):
        url = f'https://api.bilibili.com/x/space/acc/info?mid={uid}'
        return await self.get(url, headers=self.default_headers)

    async def get_user_dynamics(self, uid):
        # need_top: {1: 带置顶, 0: 不带置顶}
        url = f'https://api.vc.bilibili.com/dynamic_svr/v1/dynamic_svr/space_history?host_uid={uid}&offset_dynamic_id=0&need_top=0'
        return await self.get(url, headers=self.default_headers)
    
    async def get_new_dynamics(self):
        url = 'https://api.vc.bilibili.com/dynamic_svr/v1/dynamic_svr/dynamic_new'
        params = {'type_list': 268435455, 'access_key': self.login['access_token']}
        return await self.get(url, params=params)

    async def get_history_dynamics(self, offset_id):
        url = 'https://api.vc.bilibili.com/dynamic_svr/v1/dynamic_svr/dynamic_history'
        params = {
            'type_list': 268435455,
            'access_key': self.login['access_token'],
            'offset_dynamic_id': offset_id
        }
        return await self.get(url, params=params)
    
    async def get_live_list(self, uids):
        """根据 UID 获取直播间信息列表"""

        url = 'https://api.live.bilibili.com/room/v1/Room/get_status_info_by_uids'
        data = {'uids': uids}
        return await self.post(url, headers=self.default_headers, data=json.dumps(data))
    
    async def get_is_live_list(self):
        url = 'https://api.live.bilibili.com/xlive/app-interface/v1/relation/liveAnchor'
        params = {'access_key': self.login['access_token']}
        return await self.get(url, params=params)
    
    async def get_live_info(self, uid):
        url = f'https://api.live.bilibili.com/room/v1/Room/getRoomInfoOld?mid={uid}'
        return await self.get(url, headers=self.default_headers)

    def _sign(self, params):
        """获取 params 的 sign 值"""

        items = sorted(params.items())
        return md5(
            f"{urlencode(items)}{self.appsec}".encode('utf-8')
            ).hexdigest()

    async def get_qr(self):
        """QR码 登录"""

        params = {
            'appkey': self.appkey,
            'local_id': 0,
            'ts': int(time.time())
        }
        params['sign'] = self._sign(params)
        url = "http://passport.bilibili.com/x/passport-tv-login/qrcode/auth_code"
        r = await self.post(url, params=params)
        self.auth_code = r['auth_code']

        qr = qrcode.QRCode()
        qr.add_data(r['url'])
        img = qr.make_image()
        buf = io.BytesIO()
        img.save(buf, format='PNG')
        heximage = base64.b64encode(buf.getvalue())
        self.qr_img = heximage.decode()
        return self.qr_img

    async def qr_login(self):
        params = {
            'appkey': self.appkey,
            'local_id': 0,
            'auth_code': self.auth_code,
            'ts': int(time.time())
        }
        params['sign'] = self._sign(params)
        url = "http://passport.bilibili.com/x/passport-tv-login/qrcode/poll"
        while True:
            try:
                tokens = await self.post(url, params=params)
                Config.update_login(tokens)
                return "登入成功"
            except RequestError as e:
                if e.code == 86038:
                    return "二维码已失效，请重新登录"
                await asyncio.sleep(1)



import asyncio
import base64
import io
from urllib.parse import urlsplit

import httpx
import qrcode
from nonebot import on_command
from nonebot.adapters.cqhttp import Bot, Event
from nonebot.permission import (GROUP_ADMIN, GROUP_OWNER, PRIVATE_FRIEND,
                                SUPERUSER)
from nonebot.rule import to_me

login = on_command('登录', rule=to_me(), permission=SUPERUSER, 
    priority=5)

@login.handle()
async def _(bot: Bot, event: Event, state: dict):
    user = User()
    qr = await user.login()
    await login.send(qr)
    while True:
        if await user.login_check():
            await login.send(str(user.cookies))
            break
        await asyncio.sleep(3)

class User():
    def __init__(self):
        self.DEFAULT_HEADERS = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/\
            537.36 (KHTML, like Gecko) Chrome/79.0.3945.130 Safari/537.36",
        "Referer": "https://www.bilibili.com/"
    }

    async def get(self, url):
        async with httpx.AsyncClient() as client:
            r = await client.get(url, headers=self.DEFAULT_HEADERS)
        self.cookies = dict(r.cookies)
        r.encoding = 'utf-8'
        return r.json()
    
    async def post(self, url, data=''):
        async with httpx.AsyncClient() as client:
            r = await client.post(url, headers=self.DEFAULT_HEADERS, data=data)
        self.cookies = dict(r.cookies)
        r.encoding = 'utf-8'
        return r.json()

    async def login(self):
        get_login_url = 'https://passport.bilibili.com/qrcode/getLoginUrl'
        resp = await self.get(get_login_url)
        login_url = resp['data']['url']
        self.key = resp['data']['oauthKey']

        qr = qrcode.QRCode()
        qr.add_data(login_url)
        img = qr.make_image()

        buf = io.BytesIO()
        img.save(buf, format='PNG')
        heximage = base64.b64encode(buf.getvalue())
        self.qrb64 = f'[CQ:image,file=base64://{heximage.decode()}]'
        return self.qrb64

    async def login_check(self):
        get_login_info = 'http://passport.bilibili.com/qrcode/getLoginInfo'
        data = {'oauthKey': self.key}
        resp = await self.post(get_login_info, data=data)

        if resp['status']:
            self.url = resp['data']['url']
            return True
        return False
    
    async def get_url_cookies(self):
        query = urlsplit(self.url).query
        cookies = {}
        for s in query.split('&'):
            key, value = s.split('=')
            cookies[key] = value
        self.cookies = cookies
        return self.cookies


if __name__ == "__main__":
    async def main():
        user = User()
        qrimg = await user.login()
        print(qrimg)
        while True:
            if await user.login_check():
                print(user.cookies)
                break
            await asyncio.sleep(3)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



from datetime import datetime, timedelta

from nonebot.log import logger

from ...libs.bilireq import BiliReq
from ...database import DB as Config
from ...libs.dynamic import Dynamic
from ...utils import safe_send, scheduler

last_time = {}

@scheduler.scheduled_job('interval', seconds=10, id='dynamic_sched')
async def dy_sched():
    """直播推送"""

    with Config() as config:
        uid = config.next_uid('dynamic')
        if not uid:
            return
        push_list = config.get_push_list(uid, 'dynamic')
    
    name = push_list[0]['name']
    logger.debug(f'爬取动态 {name}（{uid}）')
    b = BiliReq()
    dynamics = (await b.get_user_dynamics(uid)).get('cards', []) # 获取最近十二条动态
    # config['uid'][uid]['name'] = dynamics[0]['desc']['user_profile']['info']['uname']
    # await update_config(config)

    if len(dynamics) == 0: # 没有发过动态或者动态全删的直接结束
        return

    if uid not in last_time: # 没有爬取过这位主播就把最新一条动态时间为 last_time
        dynamic = Dynamic(dynamics[0])
        last_time[uid] = dynamic.time
        return
    
    for dynamic in dynamics[4::-1]: # 从旧到新取最近5条动态
        dynamic = Dynamic(dynamic)
        if dynamic.time > last_time[uid] and dynamic.time > datetime.now().timestamp() - timedelta(minutes=10).seconds:
            try:
                await dynamic.get_screenshot()
            except AttributeError:
                return
            await dynamic.format()

            for sets in push_list:
                await safe_send(sets['bot_id'], sets['type'], sets['type_id'], dynamic.message)
            last_time[uid] = dynamic.time



from nonebot import on_notice
from nonebot.adapters.cqhttp import Bot, GroupDecreaseNoticeEvent
from nonebot.typing import T_State

from ..database import DB as Config
from ..version import __version__


group_decrease = on_notice(priority=5)

@group_decrease.handle()
async def _(bot: Bot, event: GroupDecreaseNoticeEvent, state: T_State):
    if event.self_id == event.user_id:
        event.message_type = 'group'
        c = Config(event)
        await c.delete_push_list()


from sqlalchemy import Boolean, Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class Sub(Base):
    __tablename__ = "subs"

    type = Column(String(10), primary_key=True)
    type_id = Column(Integer(), primary_key=True)
    uid = Column(Integer(), primary_key=True)
    live = Column(Boolean())
    dynamic = Column(Boolean())
    at = Column(Boolean())
    bot_id = Column(Integer())


class User(Base):
    __tablename__ = "users"

    uid = Column(Integer(), primary_key=True)
    name = Column(String(20))


class Group(Base):
    __tablename__ = "groups"

    id = Column(Integer(), primary_key=True)
    admin = Column(Boolean())



from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ...database import DB as Config
from ...utils import permission_check, to_me


sub_list = on_command('订阅列表', aliases={'主播列表',}, rule=to_me(), priority=5)

sub_list.handle()(permission_check)

@sub_list.handle()
async def _(bot: Bot, event: MessageEvent, state: T_State):
    with Config(event) as config:
        await sub_list.finish(await config.uid_list())


from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.permission import SUPERUSER
from nonebot.adapters.cqhttp.permission import GROUP_ADMIN, GROUP_OWNER
from nonebot.typing import T_State

from ...database import DB as Config
from ...utils import to_me


permission_off = on_command('关闭权限', rule=to_me(), 
    permission=GROUP_OWNER | GROUP_ADMIN | SUPERUSER, 
    priority=5)

@permission_off.handle()
async def _(bot: Bot, event: MessageEvent, state: T_State):
    with Config(event) as config:
        msg = await config.set_permission(False)
    await permission_off.finish(msg.replace('name', '关闭权限'))


import asyncio
import traceback

from http.client import BadStatusLine
from nonebot.log import logger
from nonebot.exception import NoLogException
from pyppeteer import launch
from pyppeteer.errors import TimeoutError

from ..database import DB as Config


class Dynamic():
    def __init__(self, dynamic):
        # self.origin = json.loads(self.card['origin'])
        self.dynamic = dynamic
        # self.card = json.loads(dynamic['card'])
        # self.dynamic['card'] = self.card
        self.type = dynamic['desc']['type']
        self.id = dynamic['desc']['dynamic_id']
        self.url = "https://t.bilibili.com/" + str(self.id)
        self.time = dynamic['desc']['timestamp']
        # self.origin_id = dynamic['desc']['orig_dy_id']
        self.uid = dynamic['desc']['user_profile']['info']['uid']
        self.name = dynamic['desc']['user_profile']['info'].get('uname', Config.get_name(self.uid))

    async def format(self):
        type_msg = {
            0: "发布了新动态",
            1: "转发了一条动态",
            8: "发布了新投稿",
            16: "发布了短视频",
            64: "发布了新专栏",
            256: "发布了新音频"
        }
        self.message = f"{self.name}{type_msg.get(self.type, type_msg[0])}：" +\
            f"\n\n传送门→{self.url}[CQ:image,file={self.img}]\n"

    async def get_screenshot(self, retry=3):
        browser = await launch(args=['--no-sandbox'])
        page = await browser.newPage()
        for i in range(retry + 1):
            try:
                await page.goto(self.url, waitUntil="networkidle0", timeout=10000)
                await page.setViewport(viewport={'width': 2560, 'height': 1080})
                card = await page.querySelector(".card")
                clip = await card.boundingBox()
                bar = await page.querySelector(".text-bar")
                bar_bound = await bar.boundingBox()
                clip['height'] = bar_bound['y'] - clip['y']
                self.img = "base64://" +\
                    await page.screenshot(clip=clip, encoding='base64')
                break
            except TimeoutError as e:
                logger.error(f"截图失败（连接超时） 已重试（{i}/{retry}）")
            except BadStatusLine as e:
                logger.error(f"截图失败（连接错误） 已重试（{i}/{retry}）")
            except Exception as e:
                logger.error("截图失败（未知错误），以下为错误日志")
                logger.error(traceback(e))
                await browser.close()
                raise
            finally:
                if i == retry:
                    try:
                        getattr(self, 'img')
                    except AttributeError:
                        logger.error("已达到重试上限，将在下个轮询中重新尝试")
                        await browser.close()
                        raise
            await asyncio.sleep(0.1)
        await browser.close()



import nonebot
from nonebot.log import logger

from ...database import DB as Config
from ...utils import safe_send, scheduler
from ...libs.bilireq import BiliReq


status = {}

@scheduler.scheduled_job('interval', seconds=10, id='live_sched')
async def live_sched():
    """直播推送"""

    with Config() as config:
        uids = config.get_uid_list('live')
        if not uids:
            return
        logger.debug(f'爬取直播列表，目前开播{sum(status.values())}人，总共{len(status)}人')
        br = BiliReq()
        r = await br.get_live_list(uids)
        if not r:
            return
        for uid, info in r.items():
            new_status = 0 if info['live_status'] == 2 else info['live_status']
            if uid not in status:
                status[uid] = new_status
                continue
            old_status = status[uid]
            if new_status != old_status and new_status: # 判断是否推送过
                room_id = info['short_id'] if info['short_id'] else info['room_id']
                url = 'https://live.bilibili.com/' + str(room_id)
                name = info['uname']
                title = info['title']
                cover = info['cover_from_user'] if info['cover_from_user'] else info['keyframe']
                push_list = config.get_push_list(uid, 'live')

                live_msg = f"{name} 开播啦！\n\n{title}\n传送门→{url}\n[CQ:image,file={cover}]\n"
                for sets in push_list:
                    at_msg = '[CQ:at,qq=all] ' if sets['at'] else ''
                    await safe_send(sets['bot_id'], sets['type'], sets['type_id'], at_msg + live_msg)
            status[uid] = new_status



from pathlib import Path

import nonebot
from nonebot import get_driver

from .config import Config

try:
    global_config = get_driver().config
    config = Config(**global_config.dict())
    from . import plugins
    # _sub_plugins = set()
    # _sub_plugins |= nonebot.load_plugins(
    #     str((Path(__file__).parent / "plugins").resolve()))
except ValueError:
    pass

from .version import VERSION, __version__



from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ...database import DB as Config
from ...utils import permission_check, to_me


add_sub = on_command('关注', aliases={'添加主播',}, rule=to_me(), priority=5)

add_sub.handle()(permission_check)

@add_sub.handle()
async def get_args(bot: Bot, event: MessageEvent, state: T_State):
    args = str(event.message).strip()
    if args:
        state['uid'] = args

@add_sub.got('uid', prompt='请输入主播的 uid')
async def _(bot: Bot, event: MessageEvent, state: T_State):
    uid = state['uid']
    with Config(event) as config:
        await add_sub.finish(await config.add_uid(uid))


from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ..utils import to_me
from ..version import __version__


get_version = on_command('版本信息', rule=to_me(), priority=5)

@get_version.handle()
async def _(bot: Bot, event: MessageEvent, state: T_State):
    message = f"当前 HarukaBot 版本：{__version__}\n" +\
        "\n使用中遇到问题欢迎加群反馈，\n" +\
        "群号：629574472\n" +\
        "\n常见问题：https://haruka-bot.live/usage/faq.html"
    await get_version.finish(message)


import os
from pathlib import Path

import nonebot
from nonebot import require
from nonebot.adapters.cqhttp import Bot, Event, MessageEvent
from nonebot.adapters.cqhttp.exception import ActionFailed, NetworkError
from nonebot.adapters.cqhttp.permission import GROUP_ADMIN, GROUP_OWNER
from nonebot.exception import FinishedException
from nonebot.log import logger
from nonebot.permission import SUPERUSER
from nonebot.rule import Rule

from .. import config

# 更换 Chromium 下载地址为非 https 淘宝源
os.environ['PYPPETEER_DOWNLOAD_HOST'] = 'http://npm.taobao.org/mirrors'
# os.environ['PYPPETEER_DOWNLOAD_HOST'] = 'http://storage.googleapis.com'
from pyppeteer.chromium_downloader import check_chromium, download_chromium

# 检查 Chromium 是否下载
if not check_chromium():
    download_chromium()


def get_path(*other):
    """获取数据文件绝对路径"""
    if config.haruka_dir:
        dir_path = Path(config.haruka_dir).absolute()
    else:
        dir_path = Path.cwd().joinpath('data')
        # dir_path = Path.cwd().joinpath('data', 'haruka_bot')
    return str(dir_path.joinpath(*other))


async def permission_check(bot: Bot, event: MessageEvent, state: dict):
    from ..database import DB as Config
    if event.message_type == 'private':
        return
    group_id = str(event.group_id)
    with Config() as config:
        if config.get_admin(group_id):
            if not await (GROUP_ADMIN | GROUP_OWNER | SUPERUSER)(bot, event):
                await bot.send(event, '权限不足，目前只有管理员才能使用')
                raise FinishedException


def to_me():
    if config.haruka_to_me:
        from nonebot.rule import to_me
        return to_me()
    async def _to_me(bot: Bot, event: Event, state: dict):
        return True
    return Rule(_to_me)


async def safe_send(bot_id, send_type, type_id, message):
    """发送出现错误时, 尝试重新发送, 并捕获异常且不会中断运行"""
    
    try:
        bot = nonebot.get_bots()[bot_id]
    except KeyError:
        logger.error(f"推送失败，Bot ID：{bot_id} 未连接")
        return

    try:
        return await bot.call_api('send_'+send_type+'_msg', **{
        'message': message,
        'user_id' if send_type == 'private' else 'group_id': type_id
        })
    except ActionFailed as e:
        logger.error(f"推送失败（操作失败），错误信息：{e.info}")
    except NetworkError as e:
        logger.error(f"推送失败（网络错误），错误信息：{e.msg}")


scheduler = require('nonebot_plugin_apscheduler').scheduler


# bot 启动时检查 src\data\haruka_bot\ 目录是否存在
if not Path(get_path()).is_dir():
    Path(get_path()).mkdir(parents=True)



from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ..utils import to_me
from ..version import __version__


# TODO 使用 doc 代替命令名列表
func_list = [
    '主播列表', '开启权限', '关闭权限', '添加主播', '删除主播', '开启动态',
    '关闭动态', '开启直播', '关闭直播', '开启全体', '关闭全体', '版本信息']


help = on_command('帮助', rule=to_me(), priority=5)

@help.handle()
async def _(bot: Bot, event: MessageEvent, state: T_State):
    message = "DD机目前支持的功能有：\n\n"
    for name in func_list:
        message += name
        if not name.endswith(('列表', '权限', '版本信息')):
            message += " uid"
        message += '\n'
    message += "\n命令中的uid需要替换为对应主播的uid，注意是uid不是直播间id\n" + \
        "\n群聊默认开启权限，只有管理员或机器人主人才能触发指令\n" + \
        "\n所有群聊/私聊的推送都是分开的，在哪里添加就会在哪里推送"
    await help.finish(message)


from typing import List

from nonebot.adapters.cqhttp.event import GroupMessageEvent, MessageEvent
from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.orm.query import Query

from .models import Base, Group, Sub, User

# TODO 启动的时候初始化推送列表
uid_list = {'live': {'list': [], 'index': 0},
            'dynamic': {'list': [], 'index': 0}}

class DB:
    """数据库交互类，与增删改查无关的部分不应该在这里面实现"""

    engine = create_engine("sqlite:///test.db")
    Base.metadata.create_all(engine)
    Session= sessionmaker(bind=engine)

    async def __aenter__(self):
        self.session: Session = self.Session()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.session.commit()
        self.session.close()

    async def add_group(self, group_id):
        """创建群设置"""

        if self.session.query(Group).filter(Group.id == group_id).first():
            # 权限开关已经创建过了
            return
        # TODO 自定义默认权限
        group = Group(id=group_id, admin=True)
        self.session.add(group)
        self.session.commit()

    async def add_sub(self, uid, type_, type_id, bot_id, name, live=True,
                      dynamic=True, at=False) -> bool:
        """添加订阅"""

        if await self.get_sub(uid, type_, type_id):
            return False
        if not await self.get_user(uid):
            await self.add_user(uid, name)
        if type_ == 'group':
            await self.add_group(type_id)
        sub = Sub(uid=uid,
                  type=type_,
                  type_id=type_id,
                  # TODO 自定义默认动态直播全体开关
                  live=live,
                  dynamic=dynamic,
                  at=at,
                  bot_id=bot_id)
        self.session.add(sub)
        self.update_uid_list()
        self.session.commit()
        return True
    
    def _get_type_id(self, event: MessageEvent):
        return (event.group_id if isinstance(event, GroupMessageEvent)
                                       else event.user_id)

    def _handle_event(self, event: MessageEvent):
        return {
            'type_': event.message_type,
            'type_id': self._get_type_id(event),
            'bot_id': event.self_id
        }

    async def add_sub_by_event(self, uid, name, event: MessageEvent):
        return await self.add_sub(uid, name=name, **self._handle_event(event))

    async def add_user(self, uid, name):
        """添加 UP 主信息"""

        user = User(uid=uid, name=name)
        self.session.add(user)
        self.session.commit()


    async def delete_group(self, group_id) -> bool:
        """删除群设置"""

        if await self.get_sub(type_='group', type_id=group_id):
            # 当前群还有订阅，不能删除
            return False

        query = self.session.query(Group).filter(Group.id == group_id)
        query.delete()
        self.session.commit()

    async def delete_sub(self, uid, type_, type_id) -> bool:
        """删除指定订阅"""

        if not await self.get_sub(uid, type_, type_id):
            # 订阅不存在
            return False
        query = await self.get_subs(uid, type_, type_id)
        query.delete()
        self.session.commit()
        await self.delete_user(uid)
        return True

    async def delete_sub_by_event(self, uid, event: MessageEvent):
        return await self.delete_sub(uid, event.message_type,
                                     self._get_type_id(event))

    async def delete_sub_list(self, type_, type_id):
        "删除指定位置的推送列表"

        subs = await self.get_subs(type_=type_, type_id=type_id)
        uids = [sub.uid for sub in subs]
        subs.delete()
        self.session.commit()
        for uid in uids:
            await self.delete_user(uid)
        if type_ == 'group':
            await self.delete_group(type_id)

    async def delete_user(self, uid) -> bool:
        """删除 UP 主信息"""

        if await self.get_sub(uid):
            # 还存在该 UP 主订阅，不能删除
            return False
        query = self.session.query(User).filter(User.uid == uid)
        query.delete()
        self.session.commit()
        return True

    async def get_push_list(self, uid, func):
        """根据类型和 UID 获取需要推送的 QQ 列表"""
        
        return (await self.get_subs(uid, **{func: True})).all()

    async def get_sub(self, uid, type_=None, type_id=None):
        """获取指定位置的订阅信息"""

        return (await self.get_subs(uid, type_, type_id)).first()

    async def get_sub_list(self, type_, type_id) -> List:
        """获取指定位置的推送列表"""

        # TODO 把文本处理部分移到外面去
        # for c in r:
        #     message += (
        #         f"【{c['name']}】" +
        #         f"直播推送：{'开' if c['live'] else '关'}，" +
        #         f"动态推送：{'开' if c['dynamic'] else '关'}" +
        #         f"（{c['uid']}）\n"
        #     )
        return (await self.get_subs(type_=type_, type_id=type_id)).all()

    async def get_subs(self, uid=None, type_=None, type_id=None, live=None,
                       dynamic=None, at=None, bot_id=None) -> Query[Sub]:
        """获取指定的订阅数据"""

        kw = locals()
        kw['type'] = kw.pop('type_')
        filters = [getattr(Sub, key) == value for key, value in kw
                   if value != None]
        return self.session.query(Sub).filter(*filters)

    async def get_uid_list(self, func):
        """根据类型获取需要爬取的 UID 列表"""
        
        return uid_list[func]['list']

    async def get_user(self, uid: int):
        """获取 UP 主信息，没有就返回 None"""

        return self.session.query(User).filter(User.uid == uid).first()

    async def next_uid(self, func):
        """获取下一个要爬取的 UID"""
        
        func = uid_list[func]

        if func['index'] >= len(func['list']):
            func['index'] = 1
            return func['list'][0]
        else:
            func['index'] += 1
            return func['list'][func['index']]

    async def set_permission(self, group_id, switch):
        """设置指定位置权限"""

        # TODO 重构为 set_group
        group = self.session.query(Group).filter(Group.id == group_id).first()
        group.admin = switch
        self.session.commit()

    async def set_sub(self, conf, switch, uid=None, type_=None, type_id=None):
        """开关订阅设置"""

        subs = await self.get_subs(uid, type_, type_id)
        if subs.count() == 0:
            return False
        for sub in subs:
            getattr(sub, conf) = switch
        self.session.commit()
        return True

    async def update_uid_list(self):
        """更新需要推送的 UP 主列表"""

        subs = self.session.query(Sub).all()
        uid_list['live']['list'] = list(set([sub.uid for sub in subs
                                             if sub.live]))
        uid_list['dynamic']['list'] = list(set([sub.uid for sub in subs
                                                if sub.dynamic]))

    @classmethod
    async def get_name(cls, uid):
        """根据 UID 获取缓存昵称"""
        pass

    @classmethod
    async def get_login(cls):
        """获取登录信息"""
        pass

    @classmethod
    async def update_login(cls, tokens):
        """更新登录信息"""
        pass

    async def new_version(self):
        """检查是否为最新版本"""
        pass

    async def update_version(self):
        """更新版本号"""
        pass

    @classmethod
    async def update_config(cls):
        """更新数据库"""
        pass


if __name__ == "__main__":
    print(type(Sub.type == 'haha'), type(User.name == 'a'))
